/*
 * Direct3D8 vertex shaders
 * Copyright (c) 2002, TransGaming Technologies Inc.
 */

#include "config.h"

#include "winbase.h"
#include "d3d8.h"
#include "debugtools.h"

#include "d3d8_private.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define D3DFVF_TEXCOUNT(x) (((x) & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT)

/* to distinguish a handle from a FVF, we'll use D3DFVF_RESERVED0,
 * assuming that HeapAlloc-ed memory is always DWORD-aligned so
 * that the lower bit of the returned pointer is always zero */
#define VS_TO_HANDLE(x) (((DWORD)x) | D3DFVF_RESERVED0)
#define VS_FROM_HANDLE(x) ((VertexShader*)(x & ~D3DFVF_RESERVED0))
#define IS_VS(x) (x & D3DFVF_RESERVED0)
#define IS_FVF(x) (!IS_VS(x))

static const char* data_types[] = {
  "D3DVSDT_FLOAT1",
  "D3DVSDT_FLOAT2",
  "D3DVSDT_FLOAT3",
  "D3DVSDT_FLOAT4",
  "D3DVSDT_D3DCOLOR",
  "D3DVSDT_UBYTE4",
  "D3DVSDT_SHORT2",
  "D3DVSDT_SHORT4",
  "8","9","10","11","12","13","14","15"
};

static void dump_vs_declaration(DWORD tok)
{
  switch ((tok & D3DVSD_TOKENTYPEMASK) >> D3DVSD_TOKENTYPESHIFT) {
  case D3DVSD_TOKEN_NOP:
    DPRINTF(" declaration: %08lx NOP()\n", tok);
    break;
  case D3DVSD_TOKEN_STREAM:
    if (tok & D3DVSD_STREAMTESSMASK) {
      DPRINTF(" declaration: %08lx STREAM_TESS()\n", tok);
    }
    else {
      DPRINTF(" declaration: %08lx STREAM(%ld)\n", tok,
	      ((tok & D3DVSD_STREAMNUMBERMASK) >> D3DVSD_STREAMNUMBERSHIFT));
    }
    break;
  case D3DVSD_TOKEN_STREAMDATA:
    if (tok & D3DVSD_DATALOADTYPEMASK) {
      DPRINTF(" declaration: %08lx SKIP(%ld)\n", tok,
	      ((tok & D3DVSD_SKIPCOUNTMASK) >> D3DVSD_SKIPCOUNTSHIFT));
    }
    else {
      DPRINTF(" declaration: %08lx REG(%ld,%s)\n", tok,
	      ((tok & D3DVSD_VERTEXREGMASK) >> D3DVSD_VERTEXREGSHIFT),
	      data_types[(tok & D3DVSD_DATATYPEMASK) >> D3DVSD_DATATYPESHIFT]);
    }
    break;
  case D3DVSD_TOKEN_TESSELLATOR:
    DPRINTF(" declaration: %08lx TESSELLATOR\n", tok);
    break;
  case D3DVSD_TOKEN_CONSTMEM:
    DPRINTF(" declaration: %08lx CONST(%ld,%ld)\n", tok,
	    ((tok & D3DVSD_CONSTADDRESSMASK) >> D3DVSD_CONSTADDRESSSHIFT),
	    ((tok & D3DVSD_CONSTCOUNTMASK) >> D3DVSD_CONSTCOUNTSHIFT));
    break;
  case D3DVSD_TOKEN_EXT:
    DPRINTF(" declaration: %08lx EXT(???)\n", tok);
    break;
  case D3DVSD_TOKEN_END:
    DPRINTF(" declaration: %08lx END()\n", tok);
    break;
  default:
    DPRINTF(" declaration: %08lx (unknown)\n", tok);
    break;
  }
}

static UINT get_instruction_argc(DWORD tok)
{
  if ((tok & 0xfffe0000) == 0xfffe0000) return 0;
  switch (tok & D3DSI_OPCODE_MASK) {
  case D3DSIO_NOP: return 0;
  case D3DSIO_MOV: return 2;
  case D3DSIO_ADD: return 3;
  case D3DSIO_SUB: return 3;
  case D3DSIO_MAD: return 4;
  case D3DSIO_MUL: return 3;
  case D3DSIO_RCP: return 2;
  case D3DSIO_RSQ: return 2;
  case D3DSIO_DP3: return 3;
  case D3DSIO_DP4: return 3;
  case D3DSIO_MIN: return 3;
  case D3DSIO_MAX: return 3;
  case D3DSIO_SLT: return 3;
  case D3DSIO_SGE: return 3;
  case D3DSIO_EXP: return 2;
  case D3DSIO_LOG: return 2;
  case D3DSIO_LIT: return 2;
  case D3DSIO_DST: return 3;
  case D3DSIO_LRP: return 4;
  case D3DSIO_FRC: return 2;
  case D3DSIO_M4x4: return 3;
  case D3DSIO_M4x3: return 3;
  case D3DSIO_M3x4: return 3;
  case D3DSIO_M3x3: return 3;
  case D3DSIO_M3x2: return 3;
  case D3DSIO_PHASE:
    return 0;
  case D3DSIO_COMMENT:
    return (tok & D3DSI_COMMENTSIZE_MASK) >> D3DSI_COMMENTSIZE_SHIFT;
  case D3DSIO_END:
    return 0;
  default:
    FIXME("unknown arg count for opcode %ld\n", tok & D3DSI_OPCODE_MASK);
    return 0;
  }
}

static void dump_instruction(CONST DWORD* ptok)
{
  DWORD tok, reg;
  const char *op, *coi;
  int c, p, srcs = 0;
  char swz[8];

  tok = *ptok++;
  if ((tok & 0xfffe0000) == 0xfffe0000) {
    DPRINTF(" instruction: %08lx VERSION %ld.%ld\n", tok,
	    D3DSHADER_VERSION_MAJOR(tok),
	    D3DSHADER_VERSION_MINOR(tok));
    return;
  }

  coi = (tok & D3DSI_COISSUE) ? " (coissued)" : "";

  switch (tok & D3DSI_OPCODE_MASK) {
  case D3DSIO_NOP:
    DPRINTF(" instruction: %08lx NOP\n", tok);
    return;
  case D3DSIO_MOV: op = "MOV"; srcs=1; break;
  case D3DSIO_ADD: op = "ADD"; srcs=2; break;
  case D3DSIO_SUB: op = "SUB"; srcs=2; break;
  case D3DSIO_MAD: op = "MAD"; srcs=3; break;
  case D3DSIO_MUL: op = "MUL"; srcs=2; break;
  case D3DSIO_RCP: op = "RCP"; srcs=1; break;
  case D3DSIO_RSQ: op = "RSQ"; srcs=1; break;
  case D3DSIO_DP3: op = "DP3"; srcs=2; break;
  case D3DSIO_DP4: op = "DP4"; srcs=2; break;
  case D3DSIO_MIN: op = "MIN"; srcs=2; break;
  case D3DSIO_MAX: op = "MAX"; srcs=2; break;
  case D3DSIO_SLT: op = "SLT"; srcs=2; break;
  case D3DSIO_SGE: op = "SGE"; srcs=2; break;
  case D3DSIO_EXP: op = "EXP"; srcs=1; break;
  case D3DSIO_LOG: op = "LOG"; srcs=1; break;
  case D3DSIO_LIT: op = "LIT"; srcs=1; break;
  case D3DSIO_DST: op = "DST"; srcs=2; break;
  case D3DSIO_LRP: op = "LRP"; srcs=3; break;
  case D3DSIO_FRC: op = "FRC"; srcs=1; break;
  case D3DSIO_M4x4: op = "M4x4"; srcs=2; break;
  case D3DSIO_M4x3: op = "M4x3"; srcs=2; break;
  case D3DSIO_M3x4: op = "M3x4"; srcs=2; break;
  case D3DSIO_M3x3: op = "M3x3"; srcs=2; break;
  case D3DSIO_M3x2: op = "M3x2"; srcs=2; break;
  case D3DSIO_PHASE:
    DPRINTF(" instruction: %08lx PHASE\n", tok);
    return;
  case D3DSIO_COMMENT:
    DPRINTF(" instruction: %08lx COMMENT \"%.*s\"\n", tok,
	    (int)(((tok & D3DSI_COMMENTSIZE_MASK) >> D3DSI_COMMENTSIZE_SHIFT)*sizeof(DWORD)),
	    (const char*)(ptok));
    return;
  case D3DSIO_END:
    DPRINTF(" instruction: %08lx END\n", tok);
    return;
  default:
    DPRINTF(" instruction: %08lx (unknown)\n", tok);
    return;
  }

  DPRINTF(" instruction: %08lx %s%s\n", tok, op, coi);

  reg = *ptok++;
  /* FIXME: parse DSTMOD */
  swz[p=0] = 0;
  if ((reg & D3DSP_WRITEMASK_ALL) != D3DSP_WRITEMASK_ALL) {
    swz[p++] = '.';
    if (reg & D3DSP_WRITEMASK_0) swz[p++] = 'x';
    if (reg & D3DSP_WRITEMASK_1) swz[p++] = 'y';
    if (reg & D3DSP_WRITEMASK_2) swz[p++] = 'z';
    if (reg & D3DSP_WRITEMASK_3) swz[p++] = 'w';
    swz[p] = 0;
  }
  switch (reg & D3DSP_REGTYPE_MASK) {
  case D3DSPR_TEMP:
    DPRINTF("    dest reg: %08lx r%ld%s\n", reg,
	    (reg & D3DSP_REGNUM_MASK), swz);
    break;
  case D3DSPR_RASTOUT:
    switch (reg & D3DSP_REGNUM_MASK) {
    case D3DSRO_POSITION:
      DPRINTF("    dest reg: %08lx oPos%s\n", reg, swz);
      break;
    case D3DSRO_FOG:
      DPRINTF("    dest reg: %08lx oFog%s\n", reg, swz);
      break;
    case D3DSRO_POINT_SIZE:
      DPRINTF("    dest reg: %08lx oPts%s\n", reg, swz);
      break;
    default:
      DPRINTF("    dest reg: %08lx (unknown-rasterizer)\n", reg);
    }
    break;
  case D3DSPR_TEXCRDOUT:
    DPRINTF("    dest reg: %08lx oT%ld%s\n", reg,
	    (reg & D3DSP_REGNUM_MASK), swz);
    break;
  default:
    DPRINTF("    dest reg: %08lx (unknown)\n", reg);
  }

  for (c=0; c<srcs; c++) {
    reg = *ptok++;
    /* FIXME: parse ADDRESSMODE */
    /* FIXME: parse SRCMOD */
    /* FIXME: parse SWIZZLE */
    switch (reg & D3DSP_REGTYPE_MASK) {
    case D3DSPR_TEMP:
      DPRINTF("     src reg: %08lx r%ld\n", reg,
	      (reg & D3DSP_REGNUM_MASK));
      break;
    case D3DSPR_INPUT:
      DPRINTF("     src reg: %08lx v%ld\n", reg,
	      (reg & D3DSP_REGNUM_MASK));
      break;
    case D3DSPR_CONST:
      DPRINTF("     src reg: %08lx c%ld\n", reg,
	      (reg & D3DSP_REGNUM_MASK));
      break;
    case D3DSPR_ADDR:
      DPRINTF("     src reg: %08lx a%ld\n", reg,
	      (reg & D3DSP_REGNUM_MASK));
      break;
    default:
      DPRINTF("     src reg: %08lx (unknown)\n", *ptok++);
    }
  }
}

HRESULT WINAPI Direct3DDevice8_CreateVertexShader(LPDIRECT3DDEVICE8 iface,
						  CONST DWORD* pDeclaration,
						  CONST DWORD* pFunction,
						  DWORD* pHandle,
						  DWORD Usage)
{
  /* ICOM_THIS(IDirect3DDevice8Impl, iface); */
  VertexShader* vs;
  CONST DWORD* ptr;
  UINT len;

  TRACE("(%p)->(%p,%p,%p,%ld)\n", iface, pDeclaration, pFunction, pHandle, Usage);

  vs = HeapAlloc(GetProcessHeap(), 0, sizeof(VertexShader));
  vs->pDeclaration = pDeclaration;
  vs->pFunction = pFunction;
  vs->Usage = Usage;
  vs->FVF = 0;

  if (pDeclaration) {
    ptr = pDeclaration;
    len = 0;
    /* FIXME: generate FVF from the declarations
     * (needed to make the HAL fixed-function pipeline happy) */
    while (*ptr != D3DVSD_END()) {
      if (TRACE_ON(ddraw))
	dump_vs_declaration(*ptr);
      ptr++; len++;
    }
    TRACE(" declaration length=%d\n", len);
  }

  if (pFunction) {
    ptr = pFunction;
    len = 0;
    while (*ptr != D3DVS_END()) {
      UINT clen = get_instruction_argc(*ptr) + 1;
      if (TRACE_ON(ddraw))
	dump_instruction(ptr);
      ptr += clen;
      len += clen;
    }
    TRACE(" function length=%d\n", len);
  }

  *pHandle = VS_TO_HANDLE(vs);

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_SetVertexShader(LPDIRECT3DDEVICE8 iface,
					       DWORD Handle)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(0x%lx)\n", iface, Handle);
  if (IS_VS(Handle)) {
    This->vertex_shader = VS_FROM_HANDLE(Handle);
    This->vertex_fvf = This->vertex_shader->FVF;
  }
  else {
    This->vertex_shader = NULL;
    This->vertex_fvf = Handle;
  }
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_GetVertexShader(LPDIRECT3DDEVICE8 iface,
					       DWORD* pHandle)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  TRACE("(%p)->(%p)\n", iface, pHandle);
  *pHandle = This->vertex_fvf;
  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_DeleteVertexShader(LPDIRECT3DDEVICE8 iface,
						  DWORD Handle)
{
  /* ICOM_THIS(IDirect3DDevice8Impl, iface); */
  VertexShader* vs;

  TRACE("(%p)->(0x%lx)\n", iface, Handle);

  vs = VS_FROM_HANDLE(Handle);
  HeapFree(GetProcessHeap(), 0, vs);

  return D3D_OK;
}

HRESULT WINAPI Direct3DDevice8_GetStridedData(LPDIRECT3DDEVICE8 iface,
					      LPD3DDRAWPRIMITIVESTRIDEDDATA lpData)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  IDirect3DVertexBuffer8Impl* buffer =
    (IDirect3DVertexBuffer8Impl*)(This->stream[0]).data;
  HRESULT hr;
  hr = Direct3DDevice8_GetStridedDataUP(iface, lpData,
					(LPVOID)buffer->t.gbl.fpVidMem,
					This->stream[0].stride);
  /* FIXME: cache returned strided data */
  return hr;
}

HRESULT WINAPI Direct3DDevice8_GetStridedDataUP(LPDIRECT3DDEVICE8 iface,
						LPD3DDRAWPRIMITIVESTRIDEDDATA lpData,
						CONST void* vertex_data,
						UINT vertex_stride)
{
  ICOM_THIS(IDirect3DDevice8Impl, iface);
  if (!This->vertex_shader) {
    /* FVF fixed-function vertex shader */
    int i;
    UINT stride = 0;
    LPBYTE ptr = (LPBYTE)vertex_data;
    memset(lpData, 0, sizeof(*lpData));
    switch (This->vertex_fvf & D3DFVF_POSITION_MASK) {
    case 0:
      break;
    case D3DFVF_XYZ:
      lpData->position.lpvData = ptr + stride;
      lpData->position.dwStride = vertex_stride;
      stride += 3*sizeof(float);
      break;
    case D3DFVF_XYZRHW:
      lpData->position.lpvData = ptr + stride;
      lpData->position.dwStride = vertex_stride;
      stride += 4*sizeof(float);
      break;
    default:
      FIXME("can't handle vertex blending yet\n");
      break;
    }
    if (This->vertex_fvf & D3DFVF_NORMAL) {
      lpData->normal.lpvData = ptr + stride;
      lpData->normal.dwStride = vertex_stride;
      stride += 3*sizeof(float);
    }
    if (This->vertex_fvf & D3DFVF_PSIZE) {
      /* FIXME: point size */
      stride += sizeof(float);
    }
    if (This->vertex_fvf & D3DFVF_DIFFUSE) {
      lpData->diffuse.lpvData = ptr + stride;
      lpData->diffuse.dwStride = vertex_stride;
      stride += sizeof(D3DCOLOR);
    }
    if (This->vertex_fvf & D3DFVF_SPECULAR) {
      lpData->specular.lpvData = ptr + stride;
      lpData->specular.dwStride = vertex_stride;
      stride += sizeof(D3DCOLOR);
    }
    if (This->vertex_fvf & D3DFVF_TEXCOUNT_MASK) {
      for (i = 0; i < D3DDP_MAXTEXCOORD; i++) {
	lpData->textureCoords[i].lpvData = ptr + stride + 2 * sizeof(float) * i;
	lpData->textureCoords[i].dwStride = vertex_stride;
      }
      stride += 2 * D3DFVF_TEXCOUNT(This->vertex_fvf) * sizeof(float);
    }
    if (stride != vertex_stride) {
      ERR("stride inconsistency! %d != %d\n", stride, vertex_stride);
    }
  }
  else if (!This->vertex_shader->pFunction) {
    /* we should be able to do this here */
    FIXME("can't handle fixed-function vertex shaders yet\n");
  }
  else {
    /* we can't do this here */
    FIXME("can't handle programmable vertex shaders yet\n");
  }
  return D3D_OK;
}

/******************************************************************
 *              ValidateVertexShader (D3D8.@)
 */
BOOL WINAPI ValidateVertexShader(LPVOID stuff)
{
  /* this function is undocumented */
  TRACE("(%p)\n", stuff);
  return TRUE;
}
